/*
 * Transparent proxy support for Linux/iptables
 *
 * Copyright (c) 2002-2004 BalaBit IT Ltd.
 * Author: Balázs Scheidler, Krisztián Kovács
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/version.h>
#include <linux/module.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#define TPROXY_VALID_HOOKS ((1 << NF_IP_PRE_ROUTING))

#if 0
#define DEBUGP printk
#else
#define DEBUGP(args...)
#endif

static struct
{
       struct ipt_replace repl;
       struct ipt_standard entries[1];
       struct ipt_error term;
} initial_table
= { { "tproxy", TPROXY_VALID_HOOKS, 2,
      sizeof(struct ipt_standard) * 1 + sizeof(struct ipt_error),
      { [NF_IP_PRE_ROUTING] 0,
       [NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) },
      { [NF_IP_PRE_ROUTING] 0,
       [NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) },
      0, NULL, { } },
    {
           /* PRE_ROUTING */
           { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
               0,
               sizeof(struct ipt_entry),
               sizeof(struct ipt_standard),
               0, { 0, 0 }, { } },
             { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
               -NF_ACCEPT - 1 } },
#if 0	       
           /* LOCAL_OUT */
           { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
               0,
               sizeof(struct ipt_entry),
               sizeof(struct ipt_standard),
               0, { 0, 0 }, { } },
             { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
               -NF_ACCEPT - 1 } }
#endif	       
    },
    /* ERROR */
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
       0,
       sizeof(struct ipt_entry),
       sizeof(struct ipt_error),
       0, { 0, 0 }, { } },
      { { { { IPT_ALIGN(sizeof(struct ipt_error_target)), IPT_ERROR_TARGET } },
         { } },
       "ERROR"
      }
    }
};

static struct ipt_table tproxy_table = {
       .name           = "tproxy",
       .valid_hooks    = TPROXY_VALID_HOOKS,
       .lock           = RW_LOCK_UNLOCKED,
       .me             = THIS_MODULE,
       .af             = AF_INET,
};

/* The work comes in here from netfilter.c. */
static unsigned int
ipt_hook(unsigned int hook,
         struct sk_buff **pskb,
         const struct net_device *in,
         const struct net_device *out,
         int (*okfn)(struct sk_buff *))
{
        return ipt_do_table(pskb, hook, in, out, &tproxy_table);
}

static struct nf_hook_ops ipt_ops[] = {
        {
                .hook = ipt_hook,
                .pf = PF_INET,
                .hooknum = NF_IP_PRE_ROUTING,
                .priority = NF_IP_PRI_NAT_SRC + 1,
                .owner = THIS_MODULE,
        },
};

static int __init iptable_tproxy_init(void)
{
	int ret;

	ret = ipt_register_table(&tproxy_table, &initial_table.repl);
	if (ret < 0)
		return ret;

	ret = nf_register_hooks(ipt_ops, ARRAY_SIZE(ipt_ops));
	if (ret < 0)
		goto cleanup_table;

	return ret;

cleanup_table:
        ipt_unregister_table(&tproxy_table);
        return ret;
}


static void __exit iptable_tproxy_fini(void)
{
        nf_unregister_hooks(ipt_ops, ARRAY_SIZE(ipt_ops));
        ipt_unregister_table(&tproxy_table);
}

module_init(iptable_tproxy_init);
module_exit(iptable_tproxy_fini);
MODULE_LICENSE("GPL");

